/*
 * genedgefeat.c - generate a features list for a texel-based edge
 *
 * Usage:	genedgefeat [-s seed] [-f descfile] [-m mindist]
 *
 * To compile:	cc -o genedgefeat genedgefeat.c -lm
 *
 * Genedgefeat is used to generate a list of texels for experiments on multi-
 * texel based edge location judgments.  The output is then given to 
 * the program texedge, which converts the feature lists to HIPS sequences.
 * Each such movie consists of a set of texels which are given random
 * x-y starting locations and a set of texel features.  These features are
 * then interpreted by texedge to indicate such things as orientation,
 * luminance, color, motion direction, and so on.  The control of the values
 * chosen for each feature is described in the file `descfile' (which defaults
 * to the standard input).  The random number generator seed may be specified
 * and defaults to the pid.  The basic logic is as follows:
 *
 * For each texel
 *	Randomly choose the location of this texel
 *	For each texel feature
 *		Randomly choose the feature value
 *
 * The texel locations are chosen randomly in order to implement a texel
 * density-based edge cue (the description file includes a description of
 * a blurred density edge), resulting in a perceived vertical edge.  The
 * locations are subject to a minimum distance between texels (specified by
 * -m, default is zero).
 *
 * Each feature's value is also chosen randomly based on control information
 * in the description file.  In order to be completely general, the feature
 * value is chosen randomly in a fairly complicated way.  For each feature,
 * the description file describes three blurred edges.  Two of these, let's
 * call them edge A and edge B, are simply blurred feature value edges.  For
 * example, edge A might be located at position 128, blurred with a sigma of 5,
 * and have a feature value on the far left of 10, and on the far right of 20.
 * A similar set of values describe edge B.  Edge description C is a
 * `probability edge' for choosing between edges A and B.  In other words, the
 * description file might describe edge C as having location 133, blur sigma
 * of 8, with a far left probability of zero and far right probability of 1.0.
 * This means that at the left, texels will have their feature value chosen
 * from edge A, at the far right from edge B, and at location 133 each texel
 * would have a probability of 0.5 of being chosen from either A or B.
 * Finally, having randomly chosen between the two blurred edges A and B, a
 * random noise value is added to the feature value so chosen, based on yet
 * another Gaussian sigma value.
 *
 * The description file format is:
 *
 * min-x min-y max-x max-y
 * number-of-texels
 * density-edge-location density-edge-sigma density-ratio
 * number-of-features
 * <feature-1-description>
 * <feature-2-description>
 *	    .
 *	    .
 *	    .
 * <feature-n-description>
 *
 * where a feature-description consists of:
 *
 * <edge-A-description>
 * <edge-B-description>
 * <edge-C-description>
 * noise-sigma
 *
 * And all feature edge descriptions consist of:
 *
 * edge-location edge-sigma left-value right-value
 *
 * The density-edge description only gives relative density (the ratio of the
 * density on the right over the density on the left).  Its
 * cumulative Gaussian is clipped at the image edges and then scaled as
 * appropriate in order to be a true cumulative probability distribution.
 * A density-ratio of 1 is treated as a special case (no density edge),
 * resulting in a speedup.  For the feature edges, a left-value equal to the
 * right-value is a similar special case (no edge at all) and results in a
 * speedup.  Throughout, a sigma of zero is acceptable, and results in a
 * special-case speedup.
 *
 * The following description file generates an edge which has two conflicting
 * cues, density and a single feature value:
 *
 * 0 0 500 500		# 500x500 image
 * 400			# 400 texels per image
 * 250 20 2		# density edge at loc'n 250, sigma 20, twice as many
 *			# texels on the right as on the left
 * 2			# 2 texel features
 * 260 15 0 10		# the first indicates an edge at loc'n 260, sigma 15,
 *			# changing from a value of 0 at the left to 10 at the
 *			# right
 * 0 0 0 0		# dummy edge B
 * 0 0 0 0		# edge C: always choose edge A
 * 0			# no noise
 * 0 0 5 5		# the second feature indicates no edge at all, but
 *			# rather is constant, always taking the value 5
 * 0 0 0 0		# dummy edge B
 * 0 0 0 0		# edge C: always choose edge A
 * 0			# no noise
 *
 * The output format consists first of the following line:
 *
 *	number-of-texels number-of-features min-x min-y max-x max-y
 *
 * This is followed by a list of texels.  For each texel the output contains:
 *
 *	x y feature-1-value feature-2-value ... feature-n-value
 *
 * Michael Landy - 4/1/92
 */

#include <stdio.h>
#include <math.h>
#include "gauss.h"

#define edgeval(z)	((z) < -5 ? 0 : ((z) > 5 ? 1 : \
			((z) < 0 ? (cgauss[5000 - ((int) ((-1000.*(z)) + .5))])\
			: (cgauss[5000 + ((int) ((1000.*(z)) + .5))]))))

#define scaleedge(z,m,M)	((m) + (((M)-(m))*(edgeval(z))))

#define zval(x,mu,sigma)	(((x) - (mu))/(sigma))

#define randflt	(random()/rand_scale)
#define	H_2PI		6.28318530717958647692

char usage[] =
	"genedgefeat: usage: genedgefeat [-s seed] [-f descfile] [-m mindist]";
char *descfile = (char *) 0;
double rand_gauss(),rand_scale = 2147483648.0;

main(argc,argv)

int argc;
char **argv;

{
	int i,j,k,l,ilow,imid,ihigh,seed = -1,ntexel,nfeature,aflag;
	double minx,miny,maxx,maxy,den_loc,den_sigma,den_ratio,x,y,p,fval;
	double *featA_loc,*featA_sigma,*featA_min,*featA_max;
	double *featB_loc,*featB_sigma,*featB_min,*featB_max;
	double *featC_loc,*featC_sigma,*featC_min,*featC_max,*feat_noisesigma;
	double density_dist[10002],mindist=0,mindistsq=0,*savex,*savey,dx,dy;
	FILE *fp;
	char s[100];

/* deal with arguments */

	for (i=1;i<argc;i++) {
		if (argv[i][0] == '-') {
			switch(argv[i][1]) {
			case 's':	if (++i >= argc)
						perr(usage);
					seed = atoi(argv[i]);
					break;
			case 'f':	if (++i >= argc)
						perr(usage);
					descfile = argv[i];
					break;
			case 'm':	if (++i >= argc)
						perr(usage);
					mindist = atof(argv[i]);
					break;
			default:	perr(usage);
			}
		}
		else
			perr(usage);
	}
	if (seed == -1)
		seed = getpid();
	srandom(seed);
	mindistsq = mindist*mindist;

/* read description file and allocate structures */

	if (descfile) {
		if ((fp = fopen(descfile,"r")) == NULL) {
			sprintf(s,"can't open file %s",descfile);
			perr(s);
		}
	}
	else {
		fp = stdin;
		descfile = "<stdin>";
	}
	fscanf(fp,"%lf %lf %lf %lf %d",&minx,&miny,&maxx,&maxy,&ntexel);
	fscanf(fp,"%lf %lf %lf",&den_loc,&den_sigma,&den_ratio);
	fscanf(fp,"%d",&nfeature);
	if (mindist) {
		if (den_ratio != 1.0)
			fprintf(stderr,
			    "genedgefeat: density edge & mindist don't mix\n");
		if ((savex = (double *) calloc(ntexel,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((savey = (double *) calloc(ntexel,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
	}
	if (nfeature > 0) {
		if ((featA_loc = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featA_sigma = (double *) calloc(nfeature,sizeof(double)))
			== (double *) 0)
				perr("can't allocate memory");
		if ((featA_min = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featA_max = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featB_loc = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featB_sigma = (double *) calloc(nfeature,sizeof(double)))
			== (double *) 0)
				perr("can't allocate memory");
		if ((featB_min = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featB_max = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featC_loc = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featC_sigma = (double *) calloc(nfeature,sizeof(double)))
			== (double *) 0)
				perr("can't allocate memory");
		if ((featC_min = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((featC_max = (double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		if ((feat_noisesigma =
			(double *) calloc(nfeature,sizeof(double))) ==
			(double *) 0)
				perr("can't allocate memory");
		for (i=0;i<nfeature;i++)
			fscanf(fp,
			  "%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
				featA_loc+i,featA_sigma+i,featA_min+i,
				featA_max+i,featB_loc+i,featB_sigma+i,
				featB_min+i,featB_max+i,featC_loc+i,
				featC_sigma+i,featC_min+i,featC_max+i,
				feat_noisesigma+i);
	}

/* precompute cumulative distribution for the density map */

	if (den_ratio != 1.) {		/* only compute if needed */
		density_dist[0] = 0;
		for (i=0;i<=10000;i++) {
			x = minx + (maxx-minx)*i/10000.;
			if (den_sigma == 0.) {
				if (x < den_loc)
					density_dist[i+1] = density_dist[i]
						+ 1.;
				else
					density_dist[i+1] = density_dist[i]
						+ den_ratio;
			}
			else
				density_dist[i+1] = density_dist[i] +
					scaleedge(zval(x,den_loc,den_sigma),
						1.,den_ratio);
		}
		x = density_dist[10001];
		for (i=0;i<=10001;i++)
			density_dist[i] /= x;
	}

	printf("%d   %d	%f %f %f %f\n",ntexel,nfeature,minx,miny,maxx,maxy);

/* main loop, once for each texel */

	for (i=0;i<ntexel;i++) {

/* choose the location by binary search */

		l = 0;
		while(1) {
			if ((++l) % 500 == 0)
				fprintf(stderr,"texel %d tried %d times\n",
					i,l);
			y = miny + (maxy-miny)*randflt;
			if (den_ratio == 1.)
				x = minx + (maxx-minx)*randflt;
			else {
				x = randflt;
				ilow = 0;
				ihigh = 10001;
				imid = 5000;
				while (ilow != imid) {
					if (density_dist[imid] <= x)
						ilow = imid;
					else
						ihigh = imid;
					imid = ilow + ((ihigh-ilow)/2);
				}
				if (ilow >= 10001)
					x = maxx;
				else
					x = minx + ((maxx-minx)*
						((ilow/10001.) +
						((x-density_dist[ilow])/
						    ((density_dist[ilow+1]-
							density_dist[ilow]) *
							10001.))));
			}
			if (mindist && i) {
				k = -1;
				for (j=0;j<i;j++) {
					dx = savex[j] - x;
					dy = savey[j] - y;
					if ((dx*dx+dy*dy) < mindistsq) {
						k = j;
						break;
					}
				}
				if (k >= 0)
					continue;
				savex[i] = x;
				savey[i] = y;
				break;
			}
			else
				break;
		}
		printf("%f %f\n",x,y);

/* feature loop for this texel */

		for (j=0;j<nfeature;j++) {
			if (featC_min[j] == featC_max[j]) {	/* no C edge */
				if (featC_min[j] == 0.)
					aflag = 1;
				else if (featC_min[j] == 1.)
					aflag = 0;
				else
					aflag = (randflt > featC_min[j]);
			}
			else if (featC_sigma[j] == 0.) {	/* no C blur */
				if (x < featC_loc[j]) {
					if (featC_min[j] == 0.)
						aflag = 1;
					else if (featC_min[j] == 1.)
						aflag = 0;
					else
						aflag =
						    (randflt > featC_min[j]);
				}
				else {
					if (featC_max[j] == 0.)
						aflag = 1;
					else if (featC_max[j] == 1.)
						aflag = 0;
					else
						aflag =
						    (randflt > featC_max[j]);
				}
			}
			else {				/* blurred C edge */
				p = scaleedge(zval(x,featC_loc[j],
					featC_sigma[j]),featC_min[j],
					featC_max[j]);
				aflag = (randflt > p);
			}
			if (aflag) {
				if (featA_min[j] == featA_max[j])
								/* no A edge */
					fval = featA_min[j];
				else if (featA_sigma[j] == 0.) {
								/* no A blur */
					if (x < featA_loc[j])
						fval = featA_min[j];
					else
						fval = featA_max[j];
				}
				else			/* blurred A edge */
					fval = scaleedge(
						zval(x,featA_loc[j],
						featA_sigma[j]),
						featA_min[j],featA_max[j]);
			}
			else {
				if (featB_min[j] == featB_max[j])
								/* no B edge */
					fval = featB_min[j];
				else if (featB_sigma[j] == 0.) {
								/* no B blur */
					if (x < featB_loc[j])
						fval = featB_min[j];
					else
						fval = featB_max[j];
				}
				else			/* blurred B edge */
					fval = scaleedge(
						zval(x,featB_loc[j],
						featB_sigma[j]),
						featB_min[j],featB_max[j]);
			}
			if (feat_noisesigma[j])
				fval += feat_noisesigma[j]*rand_gauss();
			printf(j==0 ? "	%f" : (j%5 ? " %f" : "\n	%f"),
				fval);
		}
		if (nfeature)
			printf("\n");
	}
	return(0);
}

perr(s)

char *s;

{
	fprintf(stderr,"genedgefeat: %s\n",s);
	exit(1);
}

/*
 * rand_gauss - generate random samples from a standard normal distribution
 *
 * The method is the one outlined in algorithm #267 of the ACM.
 * (based on the Box&Muller method for generating pairs
 *  of normal deviates).
 *
 * Yoav Cohen 6/2/82
 */

double rand_gauss()

{
	static int sw = 1;
	double rand1,rand2;
	static double delta,phi;

	if (sw) {
		do rand1 = randflt; while (rand1==0);
		do rand2 = randflt; while (rand2==0);
		phi = H_2PI*rand2;
		delta = sqrt(-2*log(rand1));
		sw = 0;
		return(delta*cos(phi));
	}
	sw = 1;
	return(delta*sin(phi));
}
